/*
Copyright (c) 1991 Bell Communications Research, Inc. (Bellcore)

Permission to use, copy, modify, and distribute this material 
for any purpose and without fee is hereby granted, provided 
that the above copyright notice and this permission notice 
appear in all copies, and that the name of Bellcore not be 
used in advertising or publicity pertaining to this 
material without the specific, prior written permission 
of an authorized representative of Bellcore.  BELLCORE 
MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY 
OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", 
WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES.
*/

#include <stdio.h>
#include <ctype.h>

int linepos = 0, inspace = 0, leftmargin = 0, rightmargin, biggertext=0;
int workingleft = 0, workingright, inexcerpt = 0, insignature = 0;
int termcolumns=80;

/* A common problem, in justifying text, is figuring out how to format a line when part of it wants to be left-justified, part right-justified, and part centered, or some combination thereof.  There is no perfect solution to this problem, so this program takes an incredibly cheesy but simple way out:  voting.  For each character on the line, a point is added to the "center" score if it is supposed to be centered, and so on.  If it is inside TWO "center" environments, two points are added.  This may be the world's first implementation of justification by voting... */

int centerenv=0, leftjustenv=0, rightjustenv=0;
int centerct=0, leftjustct=0, rightjustct=0;
int UsePager = 0;
int linesused = 0;

main(argc, argv)
int argc;
char **argv;
{
    int c, i, standout=0, underline=0, atstart, negated;
    char token[50], *tok;
    char tbuf[1024], *term, standoutbuf[50], standendbuf[50], *dum, StartUnderline[50], StopUnderline[50], KS[50], KE[50];
    FILE *InputFP = stdin, *tmpfp;

    if (getenv("MM_USEPAGER")) ++UsePager;
    for (i=1; i< argc; ++i) {
        if (!strcmp(argv[i], "-p")) {
            ++UsePager;
        } else {
            /* open for input */
            tmpfp = fopen(argv[i], "r");
            if (tmpfp) InputFP = tmpfp;
        }
    }
    term = (char *) getenv("TERM");
    if (term) {
        tgetent(tbuf, term);
        dum = KS;
        tgetstr("ks", &dum);
        *dum = NULL;
        dum = KE;
        tgetstr("ke", &dum);
        *dum = NULL;
        dum = standoutbuf;
        tgetstr("so", &dum);
        *dum = NULL;
        dum = standendbuf;
        tgetstr("se", &dum);
        *dum = NULL;
        dum = StartUnderline;
        tgetstr("us", &dum);
        *dum = NULL;
        dum = StopUnderline;
        tgetstr("ue", &dum);
        *dum = NULL;
        termcolumns = tgetnum("co");
        if (termcolumns <= 0) termcolumns = 80;
    } else {
        KS[0] = NULL;
        KE[0] = NULL;
        standoutbuf[0] = NULL;
        standendbuf[0] = NULL;
        StartUnderline[0] = NULL;
        StopUnderline[0] = NULL;
        termcolumns = 80;
    }
    fputs(KS, stdout);
    rightmargin = workingright = termcolumns - 1;
    while((c = getc(InputFP)) != EOF) {
        if (c == '<') { 
            for (i=0; (c = getc(InputFP)) != '>' 
		            && c != EOF; ++i) {
                token[i] = isupper(c) ? tolower(c) : c;
            }
            if (c == EOF) break;
            token[i] = NULL;
            if (token[0] == '/') {
                negated = 1;
                tok = &token[1];
            } else {
                negated = 0;
                tok = token;
            }
            switch(tok[0]) {
                case 'b':
                    if (!strcmp(tok, "bold")) {
                        if (negated) {
                            if (--standout <= 0) controloutput(standendbuf);
                        } else {
                            controloutput(standoutbuf);
                            ++standout;
                        }
                    } else if (!strcmp(tok, "bigger")) {
                        if (negated) --biggertext; else ++biggertext;
                    }
                    break;
                case 'c':
                    if (!strcmp(tok, "center")) {
                        if (negated) --centerenv; else ++centerenv;
                    } else if (!strcmp(tok, "comment")) {
                        int commct=1;
                        while (commct > 0) {
                            while ((c = getc(InputFP)) != '<' 
                                    && c != EOF) ;
                            if (c == EOF) break;
                            for (i=0; (c = getc(InputFP)) != '>' 
                              && c != EOF; ++i) {
                                token[i] = isupper(c) ? 
                                  tolower(c) : c;
                            }
                            if (c== EOF) break;
                            token[i] = NULL;
                            if (!strcmp(token, "/comment")) --commct;
                            if (!strcmp(token, "comment")) ++commct;
                        }
                    }
                    break;
                case 'e':
                    if (!strcmp(tok, "excerpt")) {
                        atstart = !(linepos > workingleft);
                        if (negated) {
                            leftmargin -= 4;
                            rightmargin += 4;
                            --inexcerpt;
                        } else {
                            leftmargin += 4;
                            rightmargin -= 4;
                            ++inexcerpt;
                        }
                        MakeWorkingMargins();
                        if (!atstart) outputc('\n');
                    }
                    break;
                case 'f':
                    if (!strcmp(tok, "flushleft")) {
                        if (negated) --leftjustenv; else ++leftjustenv;
                    } else if (!strcmp(tok, "flushright")) {
                        if (negated) --rightjustenv; else ++rightjustenv;
                    }
                    break;
                case 'i':
                    if (!strcmp(tok, "italic")) {
                        if (negated) {
                            if (--underline <= 0) controloutput(StopUnderline);
                        } else {
                            controloutput(StartUnderline);
                            ++underline;
                        }
                    } else if (!strcmp(tok, "indent")) {
                        if (negated) {
                            leftmargin -= 4;
                        } else {
                            leftmargin += 4;
                        }
                    } else if (!strcmp(tok, "indentright")) {
                        if (negated) {
                            rightmargin += 4;
                        } else {
                            rightmargin -= 4;
                        }
                    }
                    MakeWorkingMargins();
                    break;
                case 'l':
                    if (!strcmp(tok, "lt")) {
                        outputc('<');
                    }
                    break;
                case 'n':
                    if (!strcmp(tok, "nl")) {
                        outputc('\n');
                    } else if (!strcmp(tok, "np")) {
                        outputc('\n');
                        outputc('\014');
                    }
                    break;
                case 'u':
                    if (!strcmp(tok, "underline")) {
                        if (negated) {
                            if (--underline <= 0) controloutput(StopUnderline);
                        } else {
                            controloutput(StartUnderline);
                            ++underline;
                        }
                    }
                    break;
                case 'o':
                    if (!strcmp(tok, "outdent")) {
                        if (negated) {
                            leftmargin += 4;
                        } else {
                            leftmargin -= 4;
                        }
                    } else if (!strcmp(tok, "outdentright")) {
                        if (negated) {
                            rightmargin -= 4;
                        } else {
                            rightmargin += 4;
                        }
                    }
                    MakeWorkingMargins();
                    break;                    
                case 'p':
                    if (!strcmp(tok, "paragraph")) {
                        if (negated) outputc('\n');
                        outputc('\n');
                    }
                    break;
                case 's':
                    if (!strcmp(tok, "signature")) {
                        atstart = !(linepos > workingleft);
                        if (negated) {
                            leftmargin -= 4;
                            rightmargin += 4;
                            --insignature;
                        } else {
                            leftmargin += 4;
                            rightmargin -= 4;
                            ++insignature;
                        }
                        MakeWorkingMargins();
                        if (!atstart) outputc('\n');
                    }
                    break;
                default:
                    /* Ignore all other tokens */
                    break;
            }
        } else if (c == '\n') {
            if (linepos > 0 && !inspace) {
                outputc(' ');
            }
        } else {
            outputc(c);
        }
    }
    controloutput(standendbuf);
    controloutput(StopUnderline);
    outputc('\n'); /* for good measure */
    fputs(KE, stdout);
    if (UsePager) {
        Pause();
    }
    exit(0);
}

static char OutputBuf[1000] = "";
static int PendingOutput = 0, PendingControls = 0;
static int IsControlOutput[1000];

controloutput(s)
char *s;
{
    while (*s) {
        IsControlOutput[PendingOutput] = 1;
        OutputBuf[PendingOutput++] = *s++;
        ++PendingControls;
    }
}

FlushOut() {
    int i, j;
    char NewOutputBuf[1000], *s;
    int NewPendingControls=0, NewIsControlOutput[1000], NewPendingOutput = 0;

    OutputBuf[PendingOutput] = NULL;
    if (linepos >= workingright) {
        for (i=workingright; i > workingleft+4; --i) {
            if (!IsControlOutput[i]
                 && isspace(OutputBuf[i])) {
                break;
            }
        }
        for (j=0; i<PendingOutput; ++j, ++i) {
            NewOutputBuf[j] = OutputBuf[i];
            NewIsControlOutput[j] = IsControlOutput[i];
            ++NewPendingOutput;
            if (NewIsControlOutput[j]) ++NewPendingControls;
        }
        PendingOutput -= NewPendingOutput;
        PendingControls -= NewPendingControls;
        OutputBuf[PendingOutput++] = '\n';
        OutputBuf[PendingOutput] = NULL;
    }
    if ((rightjustct > leftjustct)
         && (rightjustct > centerct)
         && (rightjustct > 0)) {
        /* right justified */
        i = termcolumns - PendingOutput - 1 + PendingControls;
        for (j=0; j<i; ++j) putc(' ', stdout);
        fputs(OutputBuf, stdout);
        ++linesused;
    } else if ((leftjustct > centerct)
                && (leftjustct > 0)) {
        /* left justified */
        for (s=OutputBuf; *s && isspace(*s); ++s) {;}
        fputs(s, stdout);
        ++linesused;
    } else if (centerct > 0) {
        /* centered */
        i = (termcolumns - PendingOutput - 1 + PendingControls) / 2;
        for (j=0; j<i; ++j) putc(' ', stdout);
        fputs(OutputBuf, stdout);
        ++linesused;
    } else {
        /* Leave indentation (margins) as-is */
        fputs(OutputBuf, stdout);
        ++linesused;
    }
    if (linesused >= 22) Pause();
    rightjustct = leftjustct = centerct = 0; /* not quite right for wrapping, sigh... */
    PendingOutput = PendingControls = linepos = 0;
    inspace = 1;
    for (j=0; j<NewPendingOutput; ++j) {
        if (NewIsControlOutput[j]) {
            IsControlOutput[PendingOutput] = 1;
            OutputBuf[PendingOutput++] = NewOutputBuf[j];
            ++PendingControls;
        } else {
            realoutputc(NewOutputBuf[j], 1);
        }
    }
}

outputc(c)
char c;
{
    realoutputc(c, 0);
}

realoutputc(c, alreadyformatted)
char c;
int alreadyformatted;
{
    int i, newinspace;

    if (c == '\n') {
        IsControlOutput[PendingOutput] = 0;
        OutputBuf[PendingOutput++] = '\n';
        FlushOut();
    } else if (c == '\t') {
        int tabpos = (linepos + 8) / 8;
        if (tabpos >= workingright) {
            IsControlOutput[PendingOutput] = 0;
            OutputBuf[PendingOutput++] = '\n';
            FlushOut();
        } else {
            int spaces = (8*tabpos) - linepos;
            while (spaces-->0) {
                IsControlOutput[PendingOutput] = 0;
                OutputBuf[PendingOutput++] = ' ';
                ++linepos;
            }
        }
    } else {
        newinspace = isspace(c);
        if (!inspace || !newinspace) {
            if (linepos == 0) {
                int i = workingleft;
                if (inexcerpt) {
                    IsControlOutput[PendingOutput] = 0;
                    OutputBuf[PendingOutput++] = '>';
                    --i;
                }
                if (insignature) {
                    IsControlOutput[PendingOutput] = 0;
                    OutputBuf[PendingOutput++] = '+';
                    --i;
                }
                while (i-->0) {
                    IsControlOutput[PendingOutput] = 0;
                    OutputBuf[PendingOutput++] = ' ';
                }
                linepos = workingleft;
            }
            if (!alreadyformatted && biggertext && !(inspace && newinspace)) {
                IsControlOutput[PendingOutput] = 0;
                OutputBuf[PendingOutput++] = '_';
                ++linepos;
            }
            inspace = newinspace;
            ++linepos;
            IsControlOutput[PendingOutput] = 0;
            OutputBuf[PendingOutput++] = c;
            leftjustct += leftjustenv;
            rightjustct += rightjustenv;
            centerct += centerenv;
            if (c == '\014') inspace = 1;
            if (linepos >= workingright) FlushOut();
        }
    }
}

MakeWorkingMargins() {
    int oldworkingleft=workingleft, i;

    workingleft = leftmargin;
    workingright = rightmargin;
    if (workingleft < 0) workingleft = 0;
    if (workingright < 0) workingright = 0;
    if (workingright > (termcolumns - 1)) workingright = (termcolumns - 1);
    if (workingleft > (workingright - 8)) workingleft = workingright -8;
    if (linepos == oldworkingleft) {
        for (i=workingleft-oldworkingleft; i > 0; --i) outputc(' ');
    }
}

Pause() {
    fputs("Press RETURN to go on\n", stdout);
    getc(stdin);
    linesused = 0;
}
